/* gInfo - Copyright (C) 2004 bakunin - Andrea Marchesini 
 *                                      <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include <editor.h>

GtkSourceBuffer *sourcebuffer = NULL;
GtkWidget *editorwin = NULL;
int editor_change = 0;
GtkWidget *editor_undo_item;
GtkWidget *editor_redo_item;
GtkWidget *editor_save_item;
GtkWidget *editor_save_button;
char *editor_output = NULL;

static GtkWidget *sourceview = NULL;
static GtkWidget *editor_undo_button;
static GtkWidget *editor_redo_button;
static GtkSourceLanguagesManager *lm;

int editor_check (gpointer);
void editor_changed (GtkWidget *, gpointer);
void editor_save_file (char *);
void editor_save_as (void);
void editor_save_ok (GtkFileSelection *, gpointer);
int editor_search (char *, GtkSourceSearchFlags, int, GtkTextIter *);
int editor_search_dialog (void);

/* Add editor in notebook page */
int
editor_notebook (GtkWidget * vbox)
{
  GtkWidget *handlebox;
  GtkWidget *toolbar;
  GtkWidget *button;
  GtkWidget *sep;
  GtkTooltips *tooltips;
  GtkWidget *sw;

  GtkSourceLanguage *lang;
  GSList *l;

  /* Toolbar */
  handlebox = gtk_handle_box_new ();
  gtk_widget_show (handlebox);
  gtk_box_pack_start (GTK_BOX (vbox), handlebox, FALSE, TRUE, 0);

  toolbar = gtk_toolbar_new ();
  gtk_widget_show (toolbar);
  gtk_container_add (GTK_CONTAINER (handlebox), toolbar);
  gtk_toolbar_set_style (GTK_TOOLBAR (toolbar), GTK_TOOLBAR_ICONS);

  button = (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-new");
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (toolbar), button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, button, "Create a new XML module", NULL);

  g_signal_connect ((gpointer) button, "clicked", G_CALLBACK (editor_new),
		    NULL);

  button = (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-open");
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (toolbar), button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, button, "Open a file", NULL);

  g_signal_connect ((gpointer) button, "clicked", G_CALLBACK (editor_open),
		    NULL);

  editor_save_button =
    (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-save");
  gtk_widget_show (editor_save_button);
  gtk_container_add (GTK_CONTAINER (toolbar), editor_save_button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, editor_save_button, "Save your XML module",
			NULL);

  g_signal_connect ((gpointer) editor_save_button, "clicked",
		    G_CALLBACK (editor_save), NULL);

  button = (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-save-as");
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (toolbar), button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, button, "Save as...", NULL);

  g_signal_connect ((gpointer) button, "clicked",
		    G_CALLBACK (editor_save_as_m), NULL);

  sep = (GtkWidget *) gtk_separator_tool_item_new ();
  gtk_widget_show (sep);
  gtk_container_add (GTK_CONTAINER (toolbar), sep);

  editor_undo_button =
    (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-undo");
  gtk_widget_show (editor_undo_button);
  gtk_container_add (GTK_CONTAINER (toolbar), editor_undo_button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, editor_undo_button, "Undo the last action",
			NULL);

  gtk_widget_set_sensitive (editor_undo_button, FALSE);
  g_signal_connect ((gpointer) editor_undo_button, "clicked",
		    G_CALLBACK (editor_undo), NULL);

  editor_redo_button =
    (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-redo");
  gtk_widget_show (editor_redo_button);
  gtk_container_add (GTK_CONTAINER (toolbar), editor_redo_button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, editor_redo_button,
			"Redo the undone action", NULL);

  gtk_widget_set_sensitive (editor_redo_button, FALSE);
  g_signal_connect ((gpointer) editor_redo_button, "clicked",
		    G_CALLBACK (editor_redo), NULL);

  sep = (GtkWidget *) gtk_separator_tool_item_new ();
  gtk_widget_show (sep);
  gtk_container_add (GTK_CONTAINER (toolbar), sep);

  button = (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-find");
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (toolbar), button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, button, "Search for text", NULL);

  g_signal_connect ((gpointer) button, "clicked", G_CALLBACK (editor_find),
		    NULL);

  sep = (GtkWidget *) gtk_separator_tool_item_new ();
  gtk_widget_show (sep);
  gtk_container_add (GTK_CONTAINER (toolbar), sep);

  button = (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-cut");
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (toolbar), button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, button, "Cut the selection", NULL);

  g_signal_connect ((gpointer) button, "clicked", G_CALLBACK (editor_cut),
		    NULL);

  button = (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-copy");
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (toolbar), button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, button, "Copy the selection", NULL);

  g_signal_connect ((gpointer) button, "clicked", G_CALLBACK (editor_copy),
		    NULL);

  button = (GtkWidget *) gtk_tool_button_new_from_stock ("gtk-paste");
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (toolbar), button);
  tooltips = gtk_tooltips_new ();
  gtk_tooltips_set_tip (tooltips, button, "Past the clipboard", NULL);

  g_signal_connect ((gpointer) button, "clicked", G_CALLBACK (editor_paste),
		    NULL);

  /* ScrolledWindow */
  sw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (sw);
  gtk_box_pack_start (GTK_BOX (vbox), sw, TRUE, TRUE, 0);

  /* SouceView  & SourceBuffer */
  lm = gtk_source_languages_manager_new ();
  l = (GSList *) gtk_source_languages_manager_get_available_languages (lm);

  while (l)
    {
      char *name;

      lang = GTK_SOURCE_LANGUAGE (l->data);

      name = gtk_source_language_get_id (lang);

      if (!strcmp (name, "XML"))
	{
	  free (name);
	  break;
	}

      free (name);
      l = l->next;
    }

  if (!l)
    return 1;

  sourcebuffer =
    gtk_source_buffer_new_with_language (GTK_SOURCE_LANGUAGE (lang));

  sourceview = gtk_source_view_new_with_buffer (sourcebuffer);
  gtk_source_view_set_show_line_numbers (GTK_SOURCE_VIEW (sourceview), TRUE);
  gtk_source_view_set_auto_indent (GTK_SOURCE_VIEW (sourceview), TRUE);
  gtk_source_buffer_set_highlight (sourcebuffer, TRUE);

  gtk_widget_grab_focus (sourceview);
  gtk_widget_show (sourceview);

  gtk_text_view_set_editable (GTK_TEXT_VIEW (sourceview), TRUE);

  gtk_container_add (GTK_CONTAINER (sw), sourceview);

  g_signal_connect ((gpointer) sourcebuffer, "changed",
		    G_CALLBACK (editor_changed), NULL);

  gtk_widget_show (editorwin);

  g_timeout_add (200, editor_check, NULL);

  return 0;
}

void
editor_undo (GtkWidget * w, gpointer dummy)
{
  if (!gtk_source_buffer_can_undo (GTK_SOURCE_BUFFER (sourcebuffer)))
    return;

  gtk_source_buffer_undo (GTK_SOURCE_BUFFER (sourcebuffer));
}

/* Undo && Redo functions */
void
editor_redo (GtkWidget * w, gpointer dummy)
{
  if (!gtk_source_buffer_can_redo (GTK_SOURCE_BUFFER (sourcebuffer)))
    return;

  gtk_source_buffer_redo (GTK_SOURCE_BUFFER (sourcebuffer));
}

void
editor_changed (GtkWidget * w, gpointer dummy)
{
  editor_change = 1;
}

/* cicled function for undo/redo and save buttons */
int
editor_check (gpointer dummy)
{
  if (!gtk_source_buffer_can_redo (GTK_SOURCE_BUFFER (sourcebuffer)))
    {
      gtk_widget_set_sensitive (editor_redo_button, FALSE);
      gtk_widget_set_sensitive (editor_redo_item, FALSE);
    }
  else
    {
      gtk_widget_set_sensitive (editor_redo_button, TRUE);
      gtk_widget_set_sensitive (editor_redo_item, TRUE);
    }

  if (!gtk_source_buffer_can_undo (GTK_SOURCE_BUFFER (sourcebuffer)))
    {
      gtk_widget_set_sensitive (editor_undo_button, FALSE);
      gtk_widget_set_sensitive (editor_undo_item, FALSE);
    }
  else
    {
      gtk_widget_set_sensitive (editor_undo_button, TRUE);
      gtk_widget_set_sensitive (editor_undo_item, TRUE);
    }

  if (editor_change)
    {
      gtk_widget_set_sensitive (editor_save_button, TRUE);
      gtk_widget_set_sensitive (editor_save_item, TRUE);
    }
  else
    {
      gtk_widget_set_sensitive (editor_save_button, FALSE);
      gtk_widget_set_sensitive (editor_save_item, FALSE);
    }

  return 1;
}

/* New file */
void
editor_new (GtkWidget * w, gpointer dummy)
{
  GtkTextIter start, end;

  if (editor_change)
    if (editor_dialog_ask_save ())
      return;

  if (editor_output)
    free (editor_output);
  editor_output = NULL;

  gtk_source_buffer_begin_not_undoable_action (sourcebuffer);

  gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (sourcebuffer), &start,
				      0);
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (sourcebuffer), &end);

  gtk_text_buffer_delete (GTK_TEXT_BUFFER (sourcebuffer), &start, &end);

  editor_change = 0;

  gtk_source_buffer_end_not_undoable_action (sourcebuffer);
}

/* Open functions */
void
editor_open (GtkWidget * w, gpointer dummy)
{
  GtkWidget *dialog;
  if (editor_change)
    if (editor_dialog_ask_save ())
      return;

  if (editor_output)
    free (editor_output);
  editor_output = NULL;

  dialog =
    gtk_file_chooser_dialog_new ("Open...", GTK_WINDOW (editorwin),
				 GTK_FILE_CHOOSER_ACTION_OPEN,
				 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				 GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

  gtk_file_chooser_set_select_multiple (GTK_FILE_CHOOSER (dialog), FALSE);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      char *f;

      f = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
      editor_open_file (f);
    }

  gtk_widget_destroy (dialog);
}

void
editor_open_file (char *f)
{
  FILE *fl;
  struct stat st;
  char *a;

  if (lstat (f, &st) || !S_ISREG (st.st_mode))
    {
      editor_dialog ("I can't open this file: %s", f);
      return;
    }

  if (!(a = (char *) malloc (sizeof (char) * (st.st_size + 1))))
    fatal ("Error: memory.");

  if (!(fl = fopen (f, "r")))
    {
      free (a);
      editor_dialog ("I can't open this file: %s", f);
      return;
    }

  fread (a, st.st_size, 1, fl);
  a[st.st_size] = 0;

  gtk_source_buffer_begin_not_undoable_action (sourcebuffer);

  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (sourcebuffer), a, -1);

  fclose (fl);

  if (strlen (f) > 4 && !strcmp (f + strlen (f) - 4, ".xml"))
    editor_set_lang (NULL, "XML");

  editor_change = 0;

  gtk_source_buffer_end_not_undoable_action (sourcebuffer);

  if (!(editor_output = strdup (f)))
    fatal ("Error: memory.");
}

/* Save functions */
void
editor_save (GtkWidget * w, gpointer dummy)
{
  if (editor_output)
    editor_save_file (editor_output);
  else
    editor_save_as ();
}

void
editor_save_as_m (GtkWidget * w, gpointer dummy)
{
  editor_save_as ();
}

void
editor_save_file (char *output)
{
  FILE *fl;
  GtkTextIter start;
  GtkTextIter end;
  char *t;

  if (!(fl = fopen (output, "w")))
    {
      editor_dialog ("I can't write on %s file.", output);
      return;
    }

  gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (sourcebuffer), &start,
				      0);
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (sourcebuffer), &end);

  t =
    gtk_text_buffer_get_slice (GTK_TEXT_BUFFER (sourcebuffer), &start, &end,
			       TRUE);

  fprintf (fl, "%s", t);
  fclose (fl);

  free (t);

  editor_change = 0;
}

void
editor_save_as (void)
{
  GtkWidget *fs;

  fs = gtk_file_selection_new ("Save As...");
  gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (fs));
  gtk_file_selection_complete (GTK_FILE_SELECTION (fs), "*.xml");

  gtk_widget_show (fs);

  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (fs)->ok_button),
		      "clicked", GTK_SIGNAL_FUNC (editor_save_ok),
		      (gpointer) fs);
  gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (fs)->ok_button),
			     "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     (gpointer) fs);
  gtk_signal_connect_object (GTK_OBJECT
			     (GTK_FILE_SELECTION (fs)->cancel_button),
			     "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     (gpointer) fs);
}

void
editor_save_ok (GtkFileSelection * fs, gpointer dummy)
{
  char *filename;

  if (editor_output)
    free (editor_output);

  if ((filename =
       (gchar *)
       gtk_file_selection_get_filename (GTK_FILE_SELECTION (dummy))))
    {

      if (strcmp (filename + strlen (filename) - 4, ".xml"))
	{
	  char a[1024];

	  snprintf (a, 1024, "%s.xml", filename);

	  if (!(editor_output = strdup (a)))
	    fatal ("Error: memory");
	  editor_save_file (a);
	}
      else
	{
	  if (!(editor_output = strdup (filename)))
	    fatal ("Error: memory");
	  editor_save_file (filename);
	}
    }
}

int
editor_dialog_ask_save (void)
{
  GtkWidget *dialog;
  GtkWidget *label;
  int ret;

  dialog = gtk_dialog_new_with_buttons ("Save your module...",
					GTK_WINDOW (editorwin),
					GTK_DIALOG_MODAL |
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_YES, GTK_RESPONSE_YES,
					GTK_STOCK_NO, GTK_RESPONSE_NO,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					NULL);

  label = gtk_label_new ("Save the changes before closing?");
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), label, FALSE,
		      FALSE, 0);

  ret = gtk_dialog_run (GTK_DIALOG (dialog));

  switch (ret)
    {
    case GTK_RESPONSE_OK:
      editor_save (NULL, NULL);
      ret = 0;
      break;

    case GTK_RESPONSE_NO:
      ret = 0;
      break;

    default:
      ret = 1;
      break;
    }

  gtk_widget_destroy (dialog);

  return ret;
}

/* Message function */
void
editor_dialog (char *str, ...)
{
  GtkWidget *dialog;
  va_list va;
  char buf[1024];

  va_start (va, str);
  vsnprintf (buf, 1024, str, va);

  dialog =
    gtk_message_dialog_new (GTK_WINDOW (editorwin),
			    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
			    GTK_MESSAGE_INFO, GTK_BUTTONS_OK, buf);

  gtk_dialog_run (GTK_DIALOG (dialog));

  gtk_widget_destroy (dialog);

  va_end (va);
}

/* Cut/Copy/Paste/Delete functions */
void
editor_cut (GtkWidget * w, gpointer dummy)
{
  gtk_text_buffer_cut_clipboard (GTK_TEXT_BUFFER (sourcebuffer),
				 gtk_clipboard_get (GDK_NONE), TRUE);
}

void
editor_copy (GtkWidget * w, gpointer dummy)
{
  gtk_text_buffer_copy_clipboard (GTK_TEXT_BUFFER (sourcebuffer),
				  gtk_clipboard_get (GDK_NONE));
}

void
editor_paste (GtkWidget * w, gpointer dummy)
{
  gtk_text_buffer_paste_clipboard (GTK_TEXT_BUFFER (sourcebuffer),
				   gtk_clipboard_get (GDK_NONE), NULL, TRUE);
}

void
editor_delete (GtkWidget * w, gpointer dummy)
{
  gtk_text_buffer_delete_selection (GTK_TEXT_BUFFER (sourcebuffer), TRUE,
				    TRUE);
}

/* Find function */
void
editor_find (GtkWidget * w, gpointer dummy)
{
  GtkWidget *dialog;
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *entry;

  GtkWidget *casesensitive;
  GtkWidget *start_doc;
  GtkWidget *backward;

  GtkSourceSearchFlags flags;
  GtkTextIter iter;
  int forward;

  dialog =
    gtk_dialog_new_with_buttons ("Find For Text...", GTK_WINDOW (editorwin),
				 GTK_DIALOG_DESTROY_WITH_PARENT,
				 GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
				 GTK_STOCK_FIND, GTK_RESPONSE_OK, NULL);

  table = gtk_table_new (0, 0, FALSE);
  gtk_table_set_row_spacings (GTK_TABLE (table), 4);
  gtk_table_set_col_spacings (GTK_TABLE (table), 4);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), table, FALSE,
		      FALSE, 0);

  label = gtk_label_new ("Search For:");
  gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, 0, 1);

  entry = gtk_entry_new ();
  gtk_table_attach_defaults (GTK_TABLE (table), entry, 1, 2, 0, 1);

  casesensitive = gtk_check_button_new_with_mnemonic ("Match case");
  gtk_table_attach_defaults (GTK_TABLE (table), casesensitive, 0, 2, 1, 2);

  start_doc = gtk_check_button_new_with_mnemonic ("Search form the _head");
  gtk_table_attach_defaults (GTK_TABLE (table), start_doc, 0, 2, 2, 3);

  backward = gtk_check_button_new_with_mnemonic ("Search _backwards");
  gtk_table_attach_defaults (GTK_TABLE (table), backward, 0, 2, 3, 4);

  gtk_widget_show_all (table);

  while (1)
    {

      switch (gtk_dialog_run (GTK_DIALOG (dialog)))
	{

	case GTK_RESPONSE_OK:
	  {
	    char *what;

	    flags =
	      GTK_SOURCE_SEARCH_VISIBLE_ONLY | GTK_SOURCE_SEARCH_TEXT_ONLY;
	    if (!gtk_toggle_button_get_active
		(GTK_TOGGLE_BUTTON (casesensitive)))
	      flags |= GTK_SOURCE_SEARCH_CASE_INSENSITIVE;

	    if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (backward)))
	      forward = 0;
	    else
	      forward = 1;

	    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (start_doc)))
	      gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER
						  (sourcebuffer), &iter, 0);
	    else
	      {
		GtkTextIter sel_bound;

		gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER
						  (sourcebuffer), &iter,
						  gtk_text_buffer_get_mark
						  (GTK_TEXT_BUFFER
						   (sourcebuffer), "insert"));

		gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER
						  (sourcebuffer), &sel_bound,
						  gtk_text_buffer_get_mark
						  (GTK_TEXT_BUFFER
						   (sourcebuffer),
						   "selection_bound"));

		if (!forward)
		  gtk_text_iter_order (&sel_bound, &iter);
		else
		  gtk_text_iter_order (&iter, &sel_bound);
	      }

	    what = (gchar *) gtk_entry_get_text (GTK_ENTRY (entry));


	    if (!editor_search (what, flags, forward, &iter))
	      {
		gtk_widget_destroy (dialog);
		return;
	      }

	    break;
	  }

	default:
	  gtk_widget_destroy (dialog);
	  return;
	}
    }
}

int
editor_search (char *what, GtkSourceSearchFlags flags, int forward,
	       GtkTextIter * iter)
{

  GtkTextIter mstart, mend;
  int found;

  if (!forward)
    found =
      gtk_source_iter_forward_search (iter, what, flags, &mstart, &mend,
				      NULL);
  else
    found =
      gtk_source_iter_backward_search (iter, what, flags, &mstart, &mend,
				       NULL);

  if (found)
    {
      gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (sourcebuffer), &mstart);
      gtk_text_buffer_move_mark_by_name (GTK_TEXT_BUFFER (sourcebuffer),
					 "selection_bound", &mend);
    }

  if (!found && editor_search_dialog () == GTK_RESPONSE_YES)
    {
      gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (sourcebuffer),
					  iter, 0);
      found = editor_search (what, flags, 0, iter);
    }

  return found;
}

int
editor_search_dialog (void)
{
  GtkWidget *dialog;
  GtkWidget *label;
  int ret;

  dialog = gtk_dialog_new_with_buttons ("Find For Text...",
					GTK_WINDOW (editorwin),
					GTK_DIALOG_MODAL |
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_YES, GTK_RESPONSE_YES,
					GTK_STOCK_NO, GTK_RESPONSE_NO, NULL);
  label =
    gtk_label_new
    ("Nessun risultato trovato.\nCerco dall'inizio del documento?");
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), label, FALSE,
		      FALSE, 0);

  ret = gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_destroy (dialog);

  return ret;
}

/* Select all function */
void
editor_select_all (GtkWidget * w, gpointer dummy)
{
  GtkTextIter start, end;

  gtk_text_buffer_get_iter_at_offset (GTK_TEXT_BUFFER (sourcebuffer), &start,
				      0);
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (sourcebuffer), &end);

  gtk_text_buffer_place_cursor (GTK_TEXT_BUFFER (sourcebuffer), &end);
  gtk_text_buffer_move_mark_by_name (GTK_TEXT_BUFFER (sourcebuffer),
				     "selection_bound", &start);
}

/* Language function */
void
editor_set_lang (GtkWidget * w, gpointer dummy)
{
  char *a = (char *) dummy;
  GSList *l;
  GtkSourceLanguage *lang;
  static int no_rec_lang = 0;

  if (!sourcebuffer)
    return;

  if (no_rec_lang)
    return;

  if (!a)
    {
      gtk_source_buffer_set_highlight (GTK_SOURCE_BUFFER (sourcebuffer),
				       FALSE);
      no_rec_lang = 1;
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (highlight_normal),
				      TRUE);
      no_rec_lang = 0;
      return;
    }

  l = (GSList *) gtk_source_languages_manager_get_available_languages (lm);

  while (l)
    {
      char *name;

      lang = GTK_SOURCE_LANGUAGE (l->data);

      name = gtk_source_language_get_id (lang);

      if (!strcmp (name, a))
	{
	  free (name);
	  break;
	}

      free (name);
      l = l->next;
    }

  gtk_source_buffer_set_highlight (GTK_SOURCE_BUFFER (sourcebuffer), TRUE);
  if (l)
    {
      gtk_source_buffer_set_language (GTK_SOURCE_BUFFER (sourcebuffer), lang);

      no_rec_lang = 1;
      gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (highlight_xml),
				      TRUE);
      no_rec_lang = 0;
    }
}

/* Preferences */
void
editor_set_preferences (int tab, int spaces, int indent, int limit,
			int limit_n)
{
  gtk_source_view_set_tabs_width (GTK_SOURCE_VIEW (sourceview), tab);
  gtk_source_view_set_insert_spaces_instead_of_tabs (GTK_SOURCE_VIEW
						     (sourceview), spaces);
  gtk_source_view_set_auto_indent (GTK_SOURCE_VIEW (sourceview), indent);

  if (limit)
    gtk_source_buffer_set_max_undo_levels (sourcebuffer, limit_n);
  else
    gtk_source_buffer_set_max_undo_levels (sourcebuffer, -1);
}

/* EOF */
