/* gInfo - Copyright (C) 2004 bakunin - Andrea Marchesini 
 *                                      <bakunin@autistici.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
# error Use configure; make; make install
#endif

#include "ginfo.h"
#include "ginfo_logo.xpm"

module *module_head = NULL;
module *module_as_default = NULL;
GtkWidget *modules_window = NULL;
GtkWidget *modules_label = NULL;

int modules_c (void);
int modules_xml (void);
void modules_graphic_init (void);
void modules_graphic_text (char *, ...);
void modules_graphic_close (void);
void modules_set_default (char *, module *);
ginfo_docs *modules_xml_check_docs (xmlDocPtr, xmlNodePtr, int *);
ginfo_tips *modules_xml_check_tips (xmlDocPtr, xmlNodePtr, int *);
module_xml_info *modules_xml_check_info (xmlDocPtr, xmlNodePtr);
int modules_xml_check_info_tag (xmlDocPtr, xmlNodePtr, char *,
				module_xml_info_font *);
void modules_xml_check_info_rec (xmlDocPtr, xmlNodePtr, module_xml_info **,
				 module_xml_info **, module_xml_info_font *);
char *modules_xml_info_trim (char *);
int modules_xml_server (char *);
void modules_xml_null (void *, const char *, ...);
char *modules_xml_server_name (void);
size_t modules_xml_server_write (void *, size_t, size_t, void *);
void modules_xml_tmp_dir_mkdir (char *);

int
modules_readdir (char *server)
{
  if (!nobootsplash)
    {
      modules_graphic_init ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      usleep (500000);
    }

  if (server && modules_xml_server (server))
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      return 1;
    }

  if (modules_c () && !nobootsplash)
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      return 1;
    }

  if (modules_xml () && !nobootsplash)
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();

      return 1;
    }


  if (!nobootsplash)
    {
      modules_graphic_close ();

      while (gtk_events_pending ())
	gtk_main_iteration ();
    }

  return 0;
}

int
modules_c (void)
{
  DIR *d;
  struct dirent *dir;
  struct stat st;
  char *a;

  if (!(d = opendir (GINFO_C_PATH_MODULES)))
    return 1;

  while ((dir = readdir (d)) != NULL)
    {
      int n;

      if (strcmp (dir->d_name + (strlen (dir->d_name) - 3), ".so"))
	continue;

      n = strlen (dir->d_name) + strlen (GINFO_C_PATH_MODULES) + 2;
      if (!(a = (char *) malloc (n * sizeof (char))))
	fatal ("Error: memory.");

      snprintf (a, n, "%s/%s", GINFO_C_PATH_MODULES, dir->d_name);

      if (!lstat (a, &st) && !S_ISDIR (st.st_mode))
	{
	  if (!modules_c_check (a))
	    {
	      if (!nobootsplash)
		modules_graphic_text ("Check module %s", a);
	    }
	  else if (!nobootsplash)
	    modules_graphic_text ("Ignore file %s", a);

	  if (!nobootsplash)
	    {
	      while (gtk_events_pending ())
		gtk_main_iteration ();
	    }
	}

      free (a);
    }

  closedir (d);

  return 0;
}

int
modules_xml (void)
{
  if (modules_xml_dir (GINFO_XML_PATH_MODULES))
    return 1;
  if (modules_xml_dir (modules_xml_tmp_dir (GINFO_TMP_DIR)))
    return 1;

  return 0;
}

int
modules_xml_dir (char *directory)
{
  DIR *d;
  struct dirent *dir;
  struct stat st;
  char *a;

  if (!(d = opendir (directory)))
    return 1;

  while ((dir = readdir (d)) != NULL)
    {
      int n;

      if (strcmp (dir->d_name + (strlen (dir->d_name) - 4), ".xml"))
	continue;

      n = strlen (dir->d_name) + strlen (directory) + 2;
      if (!(a = (char *) malloc (n * sizeof (char))))
	fatal ("Error: memory.");

      snprintf (a, n, "%s/%s", directory, dir->d_name);

      if (!lstat (a, &st) && !S_ISDIR (st.st_mode))
	{
	  if (!modules_xml_check (a))
	    {

	      if (!mainwin && !nobootsplash)
		modules_graphic_text ("Check module %s", a);
	    }
	  else if (!mainwin && !nobootsplash)
	    modules_graphic_text ("Ignore file %s", a);

	  if (!nobootsplash)
	    {
	      while (gtk_events_pending ())
		gtk_main_iteration ();
	    }
	}

      free (a);
    }

  closedir (d);

  return 0;
}

int
modules_c_check (char *a)
{
  void *lib;
  int (*func0) (void);
  char *(*func1) (void);
  int (*func2) (void);
  char **(*func3) (void);
  ginfo_tips *(*func4) (int *);
  ginfo_docs *(*func5) (int *);
  int d = 0;
  char *name = NULL;
  int genre;
  char **words = NULL;
  ginfo_tips *tooltips = NULL;
  ginfo_docs *docs = NULL;
  int n;
  module *m;

  if (!(lib = dlopen (a, RTLD_LAZY)))
    {
      msg (dlerror ());
      return 1;
    }

  if ((func0 = dlsym (lib, "module_default")))
    {
      d = func0 ();
    }

  if (!(func1 = dlsym (lib, "module_name")))
    {
      dlclose (lib);
      return 1;
    }

  if (!(name = func1 ()))
    {
      dlclose (lib);
      return 1;
    }

  if (!(func2 = dlsym (lib, "module_genre")))
    {
      dlclose (lib);
      return 1;
    }

  if (!(func3 = dlsym (lib, "module_words")))
    {
      dlclose (lib);
      return 1;
    }

  if (!(words = func3 ()))
    {
      dlclose (lib);
      return 1;
    }

  if (!(func5 = dlsym (lib, "module_docs")) || !(docs = func5 (&n)))
    {
      dlclose (lib);
      return 1;
    }

  if (!(m = (module *) malloc (sizeof (module))))
    fatal ("Error: memory.");

  m->docs = NULL;
  m->type = GINFO_C;

  if (docs_add (docs, n, m))
    {
      dlclose (lib);
      free (m);
      return 1;
    }

  if (!(m->file = strdup (a)))
    fatal ("Error: memory");

  m->handle = lib;
  m->name = name;
  m->genre = genre;
  m->words = words;

  m->next = module_head;
  module_head = m;

  if (d)
    {
      modules_set_default (a, m);
    }

  if ((func4 = dlsym (lib, "module_tooltips")) != NULL
      && (tooltips = func4 (&n)) != NULL)
    tooltips_add (tooltips, n, m);

  return 0;
}

int
modules_xml_check (char *a)
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  module *m;
  char *def = NULL;
  char *name = NULL;
  char *genre = NULL;
  char *words = NULL;
  ginfo_tips *tips = NULL;
  ginfo_docs *docs = NULL;
  int docs_n = 0;
  int tips_n = 0;

  if (!(m = (module *) malloc (sizeof (module))))
    fatal ("Error: memory.");

  m->type = GINFO_XML;

  if (!(doc = xmlParseFile (a)))
    {
      msg ("Module %s: syntax error.", a);
      free (m);
      return 1;
    }

  if (!(cur = xmlDocGetRootElement (doc)))
    {
      msg ("Module %s: syntax error.", a);
      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  for (cur = cur->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "module"))
	{
	  def = xmlGetProp (cur, "default");

	  if (!(name = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*name)
	    {
	      msg ("Module %s: no name.", a);
	      free (m);
	      xmlFreeDoc (doc);
	      return 1;
	    }

	  name = trim (name);

	  if (def)
	    def = trim (def);
	}

      else if (!xmlStrcmp (cur->name, "genre"))
	{
	  if (!(genre = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*genre)
	    {
	      msg ("Module %s: no genre.", a);
	      free (m);
	      xmlFreeDoc (doc);
	      return 1;
	    }

	  genre = trim (genre);
	}

      else if (!xmlStrcmp (cur->name, "words"))
	{
	  if (!(words = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
	      || !*words)
	    {
	      msg ("Module %s: no words.", a);
	      free (m);
	      xmlFreeDoc (doc);
	      return 1;
	    }

	  words = trim (words);
	}

      else if (!xmlStrcmp (cur->name, "tooltips"))
	{
	  if (!(tips = modules_xml_check_tips (doc, cur, &tips_n)))
	    {
	      msg ("Module %s: tips syntax error.", a);
	      free (m);
	      xmlFreeDoc (doc);
	      return 1;
	    }
	}

      else if (!xmlStrcmp (cur->name, "docs"))
	{
	  if (!(docs = modules_xml_check_docs (doc, cur, &docs_n)))
	    {
	      msg ("Module %s: docs syntax error.", a);
	      free (m);
	      xmlFreeDoc (doc);
	      return 1;
	    }
	}
    }

  m->handle = doc;

  if (!name)
    {
      msg ("Module %s: no name.", a);

      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  m->name = name;

  if (!genre)
    {
      msg ("Module %s: no genre.", a);

      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  if (genre && !(m->genre = ginfo_check_genre (genre)))
    {
      msg ("Module %s: no compatible genre.", a);
      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  if (!words)
    {
      msg ("Module %s: no words.", a);
      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  if (!(m->words = ginfo_check_words (words)))
    {
      msg ("Module %s: no compatible words.", a);
      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  m->docs = NULL;

  if (!docs)
    {
      msg ("Module %s: no docs.", a);
      free (m->words);
      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  if (docs_add (docs, docs_n, m))
    {
      msg ("Module %s: no docs.", a);
      ginfo_easy_docs_list_free (docs);
      free (m->words);
      free (m);
      xmlFreeDoc (doc);
      return 1;
    }

  if (tips)
    tooltips_add (tips, tips_n, m);

  if (!(m->file = strdup (a)))
    fatal ("Error: memory");

  m->next = module_head;
  module_head = m;

  if (def && *def == '1')
    modules_set_default (a, m);

  return 0;
}

void
modules_set_default (char *a, module * m)
{
  if (module_as_default)
    {
      msg ("Error: more one module_as_default in this session: %s\n"
	   "My module_as_default is: %s\n", a, module_as_default->file);
    }
  else
    {
      module_as_default = m;
    }

}

void
modules_graphic_init (void)
{

  GtkWidget *vbox;
  GdkPixmap *pixmap;
  GdkPixmap *mask;
  GtkWidget *image;

  modules_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (modules_window), PACKAGE);
  gtk_window_set_position (GTK_WINDOW (modules_window), GTK_WIN_POS_CENTER);
  gtk_window_set_resizable (GTK_WINDOW (modules_window), FALSE);
  gtk_widget_realize (modules_window);
  gtk_widget_set_size_request (modules_window, 200, 200);


  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (modules_window), vbox);

  pixmap = gdk_pixmap_create_from_xpm_d (modules_window->window, &mask, NULL,
					 ginfo_logo);
  image = gtk_image_new_from_pixmap (pixmap, mask);
  gtk_widget_show (image);

  gtk_box_pack_start (GTK_BOX (vbox), image, TRUE, TRUE, 0);

  modules_label = gtk_label_new ("Test module:");
  gtk_widget_show (modules_label);
  gtk_box_pack_start (GTK_BOX (vbox), modules_label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (modules_label), GTK_JUSTIFY_CENTER);

  modules_label = gtk_label_new ("initialize data");
  gtk_widget_show (modules_label);
  gtk_box_pack_start (GTK_BOX (vbox), modules_label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (modules_label), GTK_JUSTIFY_CENTER);

  gtk_widget_show (modules_window);
}

void
modules_graphic_text (char *str, ...)
{

  char buf[100];
  va_list va;
  va_start (va, str);

  vsnprintf (buf, 100, str, va);
  gtk_label_set_text (GTK_LABEL (modules_label), buf);

  va_end (va);
}

void
modules_graphic_close (void)
{
  gtk_widget_destroy (modules_window);
}


void
modules_close (void)
{
  module *t;

  t = module_head;

  while (t)
    {
      module_head = t->next;
      free (t);

      t = module_head;
    }
}

ginfo_docs *
modules_xml_check_docs (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  void *t = NULL;
  xmlNodePtr cur, cur_c;
  char *chapter;
  module_xml_info *info;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "doc"))
	{

	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_docs_init ();

	  for (cur_c = cur->children; cur_c != NULL; cur_c = cur_c->next)
	    {
	      if (!xmlStrcmp (cur_c->name, "chapter"))
		{
		  if (!
		      (chapter =
		       xmlNodeListGetString (doc, cur_c->xmlChildrenNode, 1))
		      || !*chapter)
		    {
		      ginfo_easy_docs_free (t);
		      return NULL;
		    }

		  chapter = trim (chapter);

		}
	      else if (!xmlStrcmp (cur_c->name, "info"))
		{
		  if (!(info = modules_xml_check_info (doc, cur_c->children)))
		    {
		      ginfo_easy_docs_free (t);
		      return NULL;
		    }

		}
	    }

	  ginfo_easy_docs_add (t, chapter, tutor_xml, (void *) info, id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_docs_close (t, n);
}

ginfo_tips *
modules_xml_check_tips (xmlDocPtr doc, xmlNodePtr c, int *n)
{
  xmlNodePtr cur;
  char *tip;
  void *t = NULL;
  char *id;

  for (cur = c->children; cur != NULL; cur = cur->next)
    {
      if (!xmlStrcmp (cur->name, "tip"))
	{
	  if (!
	      (tip =
	       xmlNodeListGetString (doc, cur->xmlChildrenNode, 1)) || !*tip)
	    return NULL;

	  tip = trim (tip);

	  id = xmlGetProp (cur, "id");
	  if (!id || !*id)
	    id = NULL;

	  if (!t)
	    t = ginfo_easy_tips_init ();

	  ginfo_easy_tips_add (t, tip, NULL, NULL, id);
	}
    }

  if (!t)
    return NULL;

  return ginfo_easy_tips_close (t, n);
}

module_xml_info *
modules_xml_check_info (xmlDocPtr doc, xmlNodePtr cur)
{
  module_xml_info *t = NULL;
  module_xml_info *h = NULL;
  module_xml_info_font font;

  memset (&font, 0x0, sizeof (font));
  font.scale = PANGO_SCALE_MEDIUM;

  modules_xml_check_info_rec (doc, cur, &t, &h, &font);

  return h;
}

void
modules_xml_check_info_rec (xmlDocPtr doc, xmlNodePtr cur,
			    module_xml_info ** t, module_xml_info ** h,
			    module_xml_info_font * font)
{

  module_xml_info *c;

  while (cur)
    {

      c = NULL;

      if (cur->type == XML_TEXT_NODE)
	{
	  if (!(c = (module_xml_info *) malloc (sizeof (module_xml_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;

	  c->type = INFO_TEXT;
	  c->text = xmlNodeGetContent (cur);
	  if (c->text)
	    c->text = modules_xml_info_trim (c->text);

	  if (!c->text)
	    {
	      free (c);
	      c = NULL;
	    }
	}
      else if (cur->type == XML_ELEMENT_NODE)
	{
	  if (!(c = (module_xml_info *) malloc (sizeof (module_xml_info))))
	    fatal ("Error: memory.");

	  c->next = NULL;
	  c->text = NULL;

	  if (!strcmp (cur->name, "nl"))
	    c->type = INFO_NEWLINE;

	  else if (!strcmp (cur->name, "space"))
	    c->type = INFO_SPACE;

	  else if (!strcmp (cur->name, "image"))
	    {
	      c->type = INFO_IMAGE;

	      if (!
		  (c->image =
		   xmlNodeListGetString (doc, cur->xmlChildrenNode, 1))
		  || !*c->image)
		{
		  free (c);
		  c = NULL;
		}
	    }

	  else
	    {
	      c->type = INFO_FONT;

	      memcpy (&c->font, font, sizeof (module_xml_info_font));

	      if (modules_xml_check_info_tag
		  (doc, cur, (char *) cur->name, &c->font))
		{
		  free (c);
		  c = NULL;
		}
	    }
	}

      if (c)
	{
	  if (!*h)
	    *h = c;
	  else
	    (*t)->next = c;

	  *t = c;

	  if (c->type == INFO_FONT)
	    {
	      modules_xml_check_info_rec (doc, cur->children, t, h, &c->font);
	      if (!
		  (c = (module_xml_info *) malloc (sizeof (module_xml_info))))
		fatal ("Error: memory.");

	      c->next = NULL;
	      c->text = NULL;
	      (*t)->next = c;
	      *t = c;

	      c->type = INFO_FONT;

	      memcpy (&c->font, font, sizeof (module_xml_info_font));
	    }
	}

      cur = cur->next;
    }
}

int
modules_xml_check_info_tag (xmlDocPtr doc, xmlNodePtr cur, char *tag,
			    module_xml_info_font * font)
{
  /* STYLE */
  if (!strcmp (tag, "oblique"))
    {
      font->style = PANGO_STYLE_OBLIQUE;
      return 0;
    }

  if (!strcmp (tag, "italic"))
    {
      font->style = PANGO_STYLE_ITALIC;
      return 0;
    }

  /* VARIANT */
  if (!strcmp (tag, "small_caps"))
    {
      font->variant = PANGO_VARIANT_SMALL_CAPS;
      return 0;
    }

  /* WEIGHT */
  if (!strcmp (tag, "ultralight"))
    {
      font->weight = PANGO_WEIGHT_ULTRALIGHT;
      return 0;
    }

  if (!strcmp (tag, "light"))
    {
      font->weight = PANGO_WEIGHT_LIGHT;
      return 0;
    }

  if (!strcmp (tag, "bold"))
    {
      font->weight = PANGO_WEIGHT_BOLD;
      return 0;
    }

  if (!strcmp (tag, "ultrabold"))
    {
      font->weight = PANGO_WEIGHT_ULTRABOLD;
      return 0;
    }

  if (!strcmp (tag, "heavy"))
    {
      font->weight = PANGO_WEIGHT_HEAVY;
      return 0;
    }

  /* STRETCH */
  if (!strcmp (tag, "ultracondensed"))
    {
      font->stretch = PANGO_STRETCH_ULTRA_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "extracondensed"))
    {
      font->stretch = PANGO_STRETCH_EXTRA_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "condensed"))
    {
      font->stretch = PANGO_STRETCH_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "semicondensed"))
    {
      font->stretch = PANGO_STRETCH_SEMI_CONDENSED;
      return 0;
    }

  if (!strcmp (tag, "semiexpanded"))
    {
      font->stretch = PANGO_STRETCH_SEMI_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "expanded"))
    {
      font->stretch = PANGO_STRETCH_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "extraexpanded"))
    {
      font->stretch = PANGO_STRETCH_EXTRA_EXPANDED;
      return 0;
    }

  if (!strcmp (tag, "ultraexpanded"))
    {
      font->stretch = PANGO_STRETCH_ULTRA_EXPANDED;
      return 0;
    }

  /* SCALE */
  if (!strcmp (tag, "xx_small"))
    {
      font->scale = PANGO_SCALE_XX_SMALL;
      return 0;
    }

  if (!strcmp (tag, "x_small"))
    {
      font->scale = PANGO_SCALE_X_SMALL;
      return 0;
    }

  if (!strcmp (tag, "small"))
    {
      font->scale = PANGO_SCALE_SMALL;
      return 0;
    }

  if (!strcmp (tag, "large"))
    {
      font->scale = PANGO_SCALE_LARGE;
      return 0;
    }

  if (!strcmp (tag, "x_large"))
    {
      font->scale = PANGO_SCALE_X_LARGE;
      return 0;
    }

  if (!strcmp (tag, "xx_large"))
    {
      font->scale = PANGO_SCALE_XX_LARGE;
      return 0;
    }

  /* SIZE */
  if (!strcmp (tag, "size"))
    {
      char *size;
      size = xmlGetProp (cur, "value");
      if (!size || !*size)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      font->size = atoi (size) * PANGO_SCALE;

      return 0;
    }

  /* COLOR */
  if (!strcmp (tag, "foreground"))
    {
      font->foreground = xmlGetProp (cur, "value");
      if (!font->foreground || !*font->foreground)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      return 0;
    }

  if (!strcmp (tag, "background"))
    {
      font->background = xmlGetProp (cur, "value");
      if (!font->background || !*font->background)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      return 0;
    }

  /* JUSTIFY */
  if (!strcmp (tag, "justify"))
    {
      char *justify;
      justify = xmlGetProp (cur, "value");

      if (!justify || !*justify)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      if (!strcmp (justify, "center"))
	font->justify = GTK_JUSTIFY_CENTER;
      else if (!strcmp (justify, "left"))
	font->justify = GTK_JUSTIFY_LEFT;
      else if (!strcmp (justify, "right"))
	font->justify = GTK_JUSTIFY_RIGHT;
      else if (!strcmp (justify, "fill"))
	font->justify = GTK_JUSTIFY_FILL;
      else
	{
	  msg ("Axpect tag %s:value center|left|center", tag);
	  return 1;
	}

      return 0;
    }

  /* UNDERLINE */
  if (!strcmp (tag, "underline"))
    {
      font->underline = 1;
      return 0;
    }

  if (!strcmp (tag, "double_underline"))
    {
      font->underline = 2;
      return 0;
    }

  /* FACE */
  if (!strcmp (tag, "face"))
    {
      font->face = xmlGetProp (cur, "value");
      if (!font->face || !*font->face)
	{
	  msg ("Axpect tag %s value=???", tag);
	  return 1;
	}

      return 0;
    }

  if (!strcmp (tag, "font"))
    {
      char *tmp;
      tmp = xmlGetProp (cur, "style");

      /* Style */
      if (tmp)
	{
	  if (!strcmp (tmp, "oblique"))
	    font->style = PANGO_STYLE_OBLIQUE;
	  else if (!strcmp (tmp, "italic"))
	    font->style = PANGO_STYLE_ITALIC;
	  else
	    {
	      msg ("Font:Style axpected oblique|italic");
	      return 1;
	    }
	}

      /* Variant */
      tmp = xmlGetProp (cur, "variant");
      if (tmp)
	{
	  if (!strcmp (tmp, "small_cups"))
	    font->variant = PANGO_VARIANT_SMALL_CAPS;
	  else
	    {
	      msg ("Font:Variant axpected small_caps");
	      return 1;
	    }
	}

      /* Weight */
      tmp = xmlGetProp (cur, "weight");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultralight"))
	    font->weight = PANGO_WEIGHT_ULTRALIGHT;
	  else if (!strcmp (tmp, "light"))
	    font->weight = PANGO_WEIGHT_LIGHT;
	  else if (!strcmp (tmp, "bold"))
	    font->weight = PANGO_WEIGHT_BOLD;
	  else if (!strcmp (tmp, "ultrabold"))
	    font->weight = PANGO_WEIGHT_ULTRABOLD;
	  else if (!strcmp (tmp, "heavy"))
	    font->weight = PANGO_WEIGHT_HEAVY;
	  else
	    {
	      msg
		("Font:Weight axpected ultralight|light|bold|ultrabold|heavy");
	      return 1;
	    }
	}

      /* Stretch */
      tmp = xmlGetProp (cur, "stretch");
      if (tmp)
	{
	  if (!strcmp (tmp, "ultracondensed"))
	    font->stretch = PANGO_STRETCH_ULTRA_CONDENSED;
	  else if (!strcmp (tmp, "extracondensed"))
	    font->stretch = PANGO_STRETCH_EXTRA_CONDENSED;
	  else if (!strcmp (tmp, "condensed"))
	    font->stretch = PANGO_STRETCH_CONDENSED;
	  else if (!strcmp (tmp, "semicondensed"))
	    font->stretch = PANGO_STRETCH_SEMI_CONDENSED;
	  else if (!strcmp (tmp, "semiexpanded"))
	    font->stretch = PANGO_STRETCH_SEMI_EXPANDED;
	  else if (!strcmp (tmp, "expanded"))
	    font->stretch = PANGO_STRETCH_EXPANDED;
	  else if (!strcmp (tmp, "extraexpanded"))
	    font->stretch = PANGO_STRETCH_EXTRA_EXPANDED;
	  else if (!strcmp (tmp, "ultraexpanded"))
	    font->stretch = PANGO_STRETCH_ULTRA_EXPANDED;
	  else
	    {
	      msg
		("Font:Stretch axpected ultracondensed|extracondensed|condensed|semicondensed|semiexpanded|expanded|extraexpanded|ultraexpanded");
	      return 1;
	    }
	}

      /* Scale */
      tmp = xmlGetProp (cur, "scale");
      if (tmp)
	{
	  if (!strcmp (tmp, "xx_small"))
	    font->scale = PANGO_SCALE_XX_SMALL;
	  if (!strcmp (tmp, "x_small"))
	    font->scale = PANGO_SCALE_X_SMALL;
	  if (!strcmp (tmp, "small"))
	    font->scale = PANGO_SCALE_SMALL;
	  if (!strcmp (tmp, "large"))
	    font->scale = PANGO_SCALE_LARGE;
	  if (!strcmp (tmp, "x_large"))
	    font->scale = PANGO_SCALE_X_LARGE;
	  if (!strcmp (tmp, "xx_large"))
	    font->scale = PANGO_SCALE_XX_LARGE;
	  else
	    {
	      msg
		("Font:Scale axpected xx_small|x_small|small|large|x_lage|xx_large");
	      return 1;
	    }
	}

      /* Size */
      tmp = xmlGetProp (cur, "size");
      if (tmp)
	{
	  font->size = atoi (tmp) * PANGO_SCALE;
	}

      /* Color */
      font->foreground = xmlGetProp (cur, "foreground");
      font->background = xmlGetProp (cur, "background");

      /* Justify */
      tmp = xmlGetProp (cur, "justify");
      if (tmp)
	{
	  if (!strcmp (tmp, "center"))
	    font->justify = GTK_JUSTIFY_CENTER;
	  else if (!strcmp (tmp, "left"))
	    font->justify = GTK_JUSTIFY_LEFT;
	  else if (!strcmp (tmp, "right"))
	    font->justify = GTK_JUSTIFY_RIGHT;
	  else
	    {
	      msg ("Font:Justify axpected center|left|right");
	      return 1;
	    }
	}

      /* Underline */
      tmp = xmlGetProp (cur, "underline");
      if (tmp)
	{
	  if (!strcmp (tmp, "single"))
	    font->underline = 1;
	  else if (!strcmp (tmp, "double"))
	    font->underline = 2;
	  else
	    {
	      msg ("Font:Underline axpected single|double");
	      return 1;
	    }
	}

      /* Face */
      font->face = xmlGetProp (cur, "face");

      return 0;
    }

  msg ("Unknow tag %s!\n", tag);
  return 1;
}

char *
modules_xml_info_trim (char *a)
{
  register int i = 0, j = 0;
  char *b;
  int k = 0, len;

  if (!a || !*a)
    return NULL;

  len = strlen (a);

  if (!(b = (char *) malloc (sizeof (char) * (len + 1))))
    fatal ("Error: memory.");

  while (i < len)
    {
      if (a[i] == ' ' || a[i] == '\t' || a[i] == '\r' || a[i] == '\n')
	{
	  if (!k)
	    {
	      b[j++] = ' ';
	      k = 1;
	    }
	}
      else
	{
	  b[j++] = a[i];
	  k = 0;
	}

      i++;
    }

  b[j] = 0;

  for (i = k = 0; i < j; i++)
    if (b[i] != ' ')
      k++;

  if (!k)
    {
      free (b);
      return NULL;
    }

  return b;
}

/* Download remote module functions: */
char *
modules_xml_server_name (void)
{
  char *f;
  int d = 0;
  int len =
    strlen (modules_xml_tmp_dir (GINFO_TMP_DIR)) +
    strlen ("/ginfo_module_.xml") + 4;
  struct stat st;

  if (!(f = (char *) malloc (sizeof (char) * len)))
    fatal ("Error: memory.");

  while (d < 1000)
    {
      snprintf (f, len, "%s/ginfo_module_%d.xml",
		modules_xml_tmp_dir (GINFO_TMP_DIR), d);

      if (lstat (f, &st))
	break;

      d++;
    }

  if (d == 1000)
    {
      free (f);
      return NULL;
    }

  return f;
}

size_t
modules_xml_server_write (void *ptr, size_t size, size_t nmemb, void *data)
{
  register int realsize = size * nmemb;
  FILE *fl = (FILE *) data;

  fwrite (ptr, realsize, 1, fl);

  return realsize;
}

void
modules_xml_null (void *fl, const char *dummy, ...)
{
  /* Be happy :) */
}

int
modules_xml_server (char *server)
{
  char *file;

  /* This is an hack :/ */
  if (!mainwin)
    {
      modules_graphic_text ("Download %s", server);

      if (!nobootsplash)
	{
	  while (gtk_events_pending ())
	    gtk_main_iteration ();
	}
    }

  if (!(file = modules_xml_download (server)))
    return 1;

  modules_xml_server_check_list (file,
				 modules_xml_server_check_name (server));

  return 0;
}

char *
modules_xml_server_check_name (char *server_remote)
{
  int len;
  char *server;

  if (!strncmp ("http://", server_remote, 7))
    server = server_remote + 7;
  else
    server = server_remote;

  len = strlen (server);

  while (len)
    {
      if (server[len] == '/')
	break;
      len--;
    }

  if (!len)
    return server_remote;

  server[len] = 0;
  return server_remote;
}

void
modules_xml_server_check_list (char *file, char *server)
{
  xmlDocPtr doc;
  FILE *fl;
  register int i = 0;
  char buf[1024];
  char *tmp;
  int len;

  xmlSetGenericErrorFunc (xmlGenericErrorContext, modules_xml_null);

  if ((doc = xmlParseFile (file)))
    {
      xmlFree (doc);
      return;
    }

  if (!(fl = fopen (file, "r")))
    return;

  while ((buf[i] = (char) getc (fl)) != EOF)
    {
      if (buf[i] == '\n')
	{
	  buf[i] = 0;

	  if (strncmp (buf, "http://", 7))
	    {

	      len = strlen (buf) + strlen (server) + 2;
	      if (!(tmp = (char *) malloc (sizeof (char) * len)))
		fatal ("Error: memory.");

	      snprintf (tmp, len, "%s/%s", server, buf);
	      modules_xml_server (tmp);

	      free (tmp);

	    }
	  else
	    modules_xml_server (buf);

	  i = 0;
	}
      else if (i < 1024)
	i++;
    }

  fclose (fl);
  unlink (file);
}

char *
modules_xml_download (char *server)
{
  CURL *curl;
  FILE *fl;
  char *file = modules_xml_server_name ();
  long code;

  if (!file)
    return NULL;

  if (!(fl = fopen (file, "w")))
    {
      free (file);
      return NULL;
    }

  curl_global_init (CURL_GLOBAL_DEFAULT);

  if (!(curl = curl_easy_init ()))
    {
      free (file);
      return NULL;
    }

  curl_easy_setopt (curl, CURLOPT_URL, server);
  curl_easy_setopt (curl, CURLOPT_WRITEFUNCTION, modules_xml_server_write);
  curl_easy_setopt (curl, CURLOPT_FILE, (void *) fl);
  curl_easy_setopt (curl, CURLOPT_TIMEOUT, 10);

  if (curl_easy_perform (curl))
    {
      curl_easy_cleanup (curl);
      curl_global_cleanup ();

      free (file);
      return NULL;
    }

  curl_easy_getinfo (curl, CURLINFO_HTTP_CODE, &code);
  if (code != 200)
    {
      curl_easy_cleanup (curl);
      curl_global_cleanup ();

      fclose (fl);
      unlink (file);

      free (file);

      return NULL;
    }

  curl_easy_cleanup (curl);
  curl_global_cleanup ();

  fclose (fl);

  return file;
}

void
modules_xml_delete (void)
{
  DIR *d;
  struct dirent *dir;
  char *tmp;

  if (!(d = opendir (modules_xml_tmp_dir (GINFO_TMP_DIR))))
    return;

  while ((dir = readdir (d)) != NULL)
    {
      int n;

      if (strncmp (dir->d_name, "ginfo_module_", strlen ("ginfo_module_")))
	continue;

      n =
	strlen (dir->d_name) + strlen (modules_xml_tmp_dir (GINFO_TMP_DIR)) +
	2;
      if (!(tmp = (char *) malloc (sizeof (char) * n)))
	fatal ("Error: memory.");

      snprintf (tmp, n, "%s/%s", modules_xml_tmp_dir (GINFO_TMP_DIR),
		dir->d_name);
      unlink (tmp);

      free (tmp);
    }
}

/* Remove functions */
module *
modules_search (char *mod)
{
  module *m = module_head;

  while (m)
    {
      if (!strcmp (m->file, mod))
	return m;

      m = m->next;
    }

  return NULL;
}

void
modules_delete (module * m_del)
{
  module *m, *h = NULL;
  tooltip *t, *k = NULL, *j;
  doc *d;

  m = module_head;

  while (m)
    {
      if (m == m_del)
	{
	  if (!h)
	    module_head = m->next;
	  else
	    h->next = m->next;

	  if (m->type == GINFO_C)
	    {
	      dlclose (m->handle);
	    }
	  else
	    {
	      xmlFree (m->handle);
	      free (m->words);
	    }

	  if (module_as_default == m)
	    module_as_default = NULL;

	  t = tooltip_head;

	  while (t)
	    {
	      if (t->module == m)
		{

		  tooltip_number--;

		  if (!k)
		    tooltip_head = t->next;
		  else
		    k->next = t->next;

		  j = t->next;
		  free (t);
		  t = j;
		}
	      else
		{
		  k = t;
		  t = t->next;
		}
	    }

	  while (m->docs)
	    {
	      module_xml_info *i, *f;
	      d = m->docs;

	      i = d->arg;
	      while (i)
		{
		  f = i;
		  i = i->next;
		  free (f);
		}

	      m->docs = m->docs->next;

	      free (d);
	    }

	  free (m);

	  break;
	}

      h = m;
      m = m->next;
    }
}

/* tmp_dir */
char *
modules_xml_tmp_dir (char *a)
{
  static char *ret = NULL;
  int len;
  char *home;
  char *tmp;

  if (ret)
    {
      modules_xml_tmp_dir_mkdir (ret);
      return ret;
    }

  if (!(home = getenv ("HOME")))
    {
      ret = a;
      modules_xml_tmp_dir_mkdir (ret);
      return ret;
    }

  len = strlen (home) + strlen ("/.ginfo") + 1;
  if (!(tmp = (char *) malloc (sizeof (char) * len)))
    fatal ("Error: memory.");

  snprintf (tmp, len, "%s/.ginfo", home);

  ret = tmp;

  modules_xml_tmp_dir_mkdir (ret);

  return ret;
}

void
modules_xml_tmp_dir_mkdir (char *f)
{
  struct stat st;

  if (lstat (f, &st))
    mkdir (f, 0750);
}

/* EOF */
